<?php

namespace com\fwdekker\deathnotifier;

use com\fwdekker\deathnotifier\validator\InvalidInputException;
use InvalidArgumentException;
use Monolog\Logger;


/**
 * Dispatches {@see Action Actions} to the right implementation.
 */
class ActionDispatcher
{
    /**
     * @var Logger the logger to log with
     */
    private readonly Logger $logger;
    /**
     * @var array<string, array<string, Action>> the registered actions
     */
    private array $actions = [];


    /**
     * Constructs a new `ActionDispatcher`.
     */
    public function __construct()
    {
        $this->logger = LoggerUtil::with_name($this::class);
    }


    /**
     * Registers {@see $action} so that {@see ActionDispatcher::handle()} can find it.
     *
     * Only one `Action` can be registered given a combination of {@see ActionMethod} and action name. An
     * {@see IllegalArgumentError} is thrown if this restriction is violated.
     *
     * @param ActionMethod $method the `ActionMethod` that {@see $action} should handle
     * @param string $action_name the action name that {@see $action} should handle
     * @param Action $action the `Action` to register
     * @return void
     */
    public function register_action(ActionMethod $method, string $action_name, Action $action): void
    {
        $method_name = $method->name;

        if (!isset($this->actions[$method_name]))
            $this->actions[$method_name] = [];

        if (isset($this->actions[$method_name][$action_name]))
            throw new IllegalArgumentError("Cannot register another handler for $method_name action '$action_name'.");

        $this->actions[$method_name][$action_name] = $action;
    }

    /**
     * Utility method that invokes {@see ActionDispatcher::register_action()} for each of {@see $actions}.
     *
     * @param array<array{ActionMethod, string, Action}> $actions the actions to register, mapped from the
     * `ActionMethod` method and action name that the `Action` should handle
     * @return void
     */
    public function register_actions(array $actions): void
    {
        foreach ($actions as $action)
            $this->register_action($action[0], $action[1], $action[2]);
    }

    /**
     * Executes the {@see Action} registered with {@see ActionDispatcher::register_action()} that matches both
     * {@see $method} and the `action` key in {@see $inputs}.
     *
     * @param ActionMethod $method the method of the action to execute
     * @param array<int|string, mixed> $inputs the inputs to the action, where the `action` key specifies which action
     * to execute
     * @return Response an unsatisfied `Response` if no registered {@see Action} could be found to handle the request,
     * an unsatisfied `Response` if an `Action` could be found but threw an {@see UnexpectedException} or an
     * {@see InvalidInputException}, or a satisfied `Response` containing the return value of `Action` that was invoked
     * with {@see $inputs}
     */
    public function handle(ActionMethod $method, array $inputs): Response
    {
        if (!isset($inputs["action"]))
            return Response::unsatisfied(message: "Malformed request: No action specified.");

        $method_name = $method->name;
        $action_name = $inputs["action"];
        if (!isset($this->actions[$method_name]) || !isset($this->actions[$method_name][$action_name]))
            return Response::unsatisfied(message: "Malformed request: Unknown $method_name action '$action_name'.");

        $action = $this->actions[$method_name][$action_name];
        try {
            return new Response(
                satisfied: true,
                payload: $action->handle($inputs)
            );
        } catch (UnexpectedException $exception) {
            $this->logger->error($exception);
            return Response::unsatisfied(message: $exception->getMessage());
        } catch (InvalidInputException $exception) {
            return Response::unsatisfied(message: $exception->getMessage(), target: $exception->target);
        }
    }
}
