<?php

namespace com\fwdekker\deathnotifier\user;

use com\fwdekker\deathnotifier\Action;
use com\fwdekker\deathnotifier\validator\IsValidCsrfTokenRule;
use com\fwdekker\deathnotifier\Util;
use com\fwdekker\deathnotifier\validator\IsEmailRule;
use com\fwdekker\deathnotifier\validator\InvalidInputException;
use com\fwdekker\deathnotifier\validator\IsStringRule;
use com\fwdekker\deathnotifier\validator\RuleSet;


/**
 * Verifies the user's email address.
 */
class VerifyEmailAction extends Action
{
    /**
     * @var UserList the list to verify the email address in
     */
    private readonly UserList $user_list;


    /**
     * Constructs a new `VerifyEmailAction`.
     *
     * @param UserList $user_list the list to verify the email address in
     */
    public function __construct(UserList $user_list)
    {
        $this->user_list = $user_list;
    }


    /**
     * Verifies the user's email address.
     *
     * Requires that a valid CSRF token is present. Does not require the user to be logged in or out.
     *
     * @param array<int|string, mixed> $inputs `"token": string`: a valid CSRF token, `"email"`: the email address to
     * verify, `"verify_token": string`: the token to verify the email address with
     * @return null
     * @throws InvalidInputException if no valid CSRF token is present, if no account with the given email address
     * exists, if the reset token is invalid, or if the reset token has expired
     */
    public function handle(array $inputs): mixed
    {
        (new RuleSet([
            "token" => [new IsValidCsrfTokenRule()],
            "email" => [new IsEmailRule()],
            "verify_token" => [new IsStringRule()],
        ]))->check($inputs);

        $this->user_list->transaction(function () use ($inputs) {
            $user_data = $this->user_list->get_user_by_email($inputs["email"]);
            if ($user_data === null)
                throw new InvalidInputException("No user with that email address has been registered.");
            if ($user_data["email_verification_token"] !== $inputs["verify_token"])
                // TODO: Just tell the user whether the account has been verified or whether the token is plain wrong
                // TODO: Also, tell the user what they can do to resolve this
                throw new InvalidInputException(
                    "Failed to verify email address. Maybe you already verified your email address?"
                );

            $minutes_left = Util::minutes_until_interval_elapsed(
                $user_data["email_verification_token_timestamp"],
                UserList::MINUTES_VALID_VERIFICATION
            );
            if ($minutes_left < 0)
                throw new InvalidInputException(
                    "This email verification link has expired. Log in and request a new verification email."
                );

            $this->user_list->set_email_verified($_SESSION["uuid"]);
        });

        return null;
    }
}
