<?php

namespace com\fwdekker\deathnotifier\validator;


/**
 * Validates that the input is of the specific length.
 */
class HasStringLengthRule extends Rule
{
    /**
     * @var int|null the minimum length (inclusive), or `null` if there is no minimum length
     */
    private readonly ?int $min_length;
    /**
     * @var int|null the maximum length (inclusive), or `null` if there is no maximum length
     */
    private readonly ?int $max_length;


    /**
     * Constructs a new `HasLengthRule`.
     *
     * @param int|null $min_length the minimum length (inclusive), or `null` if there is no minimum length
     * @param int|null $max_length the maximum length (inclusive), or `null` if there is no maximum length
     * @param string|null $override_message the message to return if the rule does not apply to some input. If `null`,
     * the rule implementation can choose an appropriate message
     */
    public function __construct(?int $min_length = null, ?int $max_length = null, ?string $override_message = null)
    {
        parent::__construct($override_message);

        $this->min_length = $min_length;
        $this->max_length = $max_length;
    }


    /**
     * Validates that the input is of the specific length.
     *
     * @param array<int|string, mixed> $inputs the list of inputs in which the value at {@see $key} should be checked
     * @param string $key the key in {@see $inputs} of the input to check
     * @return void if the checked input is of the specified length
     * @throws InvalidInputException if the checked input is not set or is not of the specified length
     */
    public function check(array $inputs, string $key): void
    {
        if (!isset($inputs[$key]))
            throw new InvalidInputException($this->override_message ?? "Missing input '$key'.", $key);

        if (!is_string($inputs[$key]))
            throw new InvalidInputException($this->override_message ?? "Field must be a string.", $key);

        if ($this->min_length !== null && strlen($inputs[$key]) < $this->min_length)
            throw new InvalidInputException(
                $this->override_message ?? "Use at least $this->min_length character(s).",
                $key
            );
        if ($this->max_length !== null && strlen($inputs[$key]) > $this->max_length)
            throw new InvalidInputException(
                $this->override_message ?? "Use at most $this->max_length character(s).",
                $key
            );
    }
}
